; docformat = 'rst'

;+
; IDL-OPENDAP interface.
;
; :Private:
;
; :Author: 
;    Patrick West <mrpatrickwest@gmail.com> UCAR/HAO
;
;    Based on IDL-DODS Written by Daniel J. Carr - Research Systems, Inc.
;
; :Copyright:
;    Copyright (c) 2006 OPeNDAP, Inc.
;
;    This is free software; you can redistribute it and/or
;    modify it under the terms of the GNU Lesser General Public
;    License as published by the Free Software Foundation; either
;    version 2.1 of the License, or (at your option) any later version.
; 
;    This library is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;    Lesser General Public License for more details.
; 
;    You should have received a copy of the GNU Lesser General Public
;    License along with this library; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;   You can contact OPeNDAP, Inc. at PO Box 112, Saunderstown, RI. 02874-0112.
;-

;+
; Handle a single node (and its children).
; 
; :Returns:
;    status (1=ok, 0=fail)
;
; :Params:
;    opendap_data : in, out, required, type=structure
;       returned data for node
;    node : in, required, type=long
;       node ID
;    get_data : in, required, type=int
;       set to 1 to get_data, 0 to not get any data
;    debug : in, required, type=int
;       set to 1 to debug
;
; :Keywords:
;    is_empty : out, optional, type=boolean
;       set to named variable to return if the node is empty
;    indicate_type : in, optional, type=boolean
;       set to have each structure in the return value indicate it's type 
;       (i.e. "GRID", "GROUP", etc.)
;-
FUNCTION opendap_handle_node, opendap_data, node, get_data, debug, $
                              IS_EMPTY=empty, INDICATE_TYPE=indicateType
  COMPILE_OPT strictarr
  
  stat = 0UL

  OPENDAP_NODE_INFO, node, node_name, node_type, num_subnodes, $
                     has_content, has_attrs, content

  OPENDAP_FIX_NAME, node_name
  IF (debug EQ 1UL) THEN BEGIN
    PRINT, 'node_name = ', node_name
    PRINT, 'node_type = ', node_type
    PRINT, 'num_subnodes = ', num_subnodes
    PRINT, 'has_content = ', has_content
    PRINT, 'has_attrs = ', has_attrs
  ENDIF

  IF (has_attrs EQ 1UL) THEN BEGIN
    stat = OPENDAP_HANDLE_ATTRIBUTES(opendap_data, node, debug, $
                                     INDICATE_TYPE=indicateType)
    IF (KEYWORD_SET(indicateType)) THEN BEGIN
      opendap_data = CREATE_STRUCT('_TYPE', 'ATTRIBUTES', opendap_data)
    ENDIF
  ENDIF ELSE BEGIN
    IF (has_content EQ 1UL) THEN BEGIN
      stat = OPENDAP_HANDLE_CONTENT(opendap_data, node, content, $
                                    node_type, has_content, get_data, $
                                    debug, NAME=node_name, $
                                    INDICATE_TYPE=indicateType)
    ENDIF ELSE BEGIN
      IF (num_subnodes GT 0UL) THEN BEGIN
        empty = 1B
        
        FOR i_num_subnodes = 0UL, num_subnodes - 1L DO BEGIN
          OPENDAP_GET_SUBNODE, node, i_num_subnodes, subnode, subnode_name
          OPENDAP_FIX_NAME, subnode_name
          stat = OPENDAP_HANDLE_NODE(opendap_sub_data, subnode, $
                                     get_data, debug, IS_EMPTY=subEmpty, $
                                     INDICATE_TYPE=indicateType)
          IF (~KEYWORD_SET(subEmpty)) THEN BEGIN
            IF (empty) THEN BEGIN
              opendap_data = CREATE_STRUCT(subnode_name, opendap_sub_data)
            ENDIF ELSE BEGIN                    
              opendap_data = CREATE_STRUCT(opendap_data, subnode_name, $
                                           opendap_sub_data)
            ENDELSE
            
            ; no longer empty
            empty = 0B            
          ENDIF
        ENDFOR

        IF (KEYWORD_SET(indicateType)) THEN BEGIN
          opendap_data = CREATE_STRUCT('_TYPE', 'GROUP', opendap_data)
        ENDIF        
      ENDIF ELSE BEGIN
        stat = OPENDAP_HANDLE_CONTENT(opendap_data, node, content, $
                                      node_type, has_content, $
                                      get_data, debug, IS_EMPTY=subEmpty, $
                                      NAME=node_name, $
                                      INDICATE_TYPE=indicateType)
        empty = KEYWORD_SET(subEmpty)
      ENDELSE
    ENDELSE
  ENDELSE

  RETURN, stat
END

